"""
quart_wtf.utils

Utilities for `quart_wtf`.
"""
import hashlib
import hmac
import logging
import os
from typing import Any, Optional, Union
from urllib.parse import urlparse
from werkzeug.datastructures import CombinedMultiDict, ImmutableMultiDict, MultiDict

from itsdangerous import BadData, SignatureExpired, URLSafeTimedSerializer
from quart import current_app, g, request, session
from wtforms import ValidationError

from .const import (CSRF_NOT_CONFIGURED, FIELD_NAME_REQUIRED, SECRET_KEY_REQUIRED,
                   SESSION_TOKEN_MISSING, TOKEN_EXPIRED, TOKEN_INVALID, TOKEN_MISSING,
                   TOKEN_NO_MATCH)


__all__ = [
    "logger",
    "SUBMIT_METHODS",
    "_is_submitted",
    "_get_formdata",
    "_get_config",
    "generate_csrf",
    "validate_csrf",
    "same_origin"
]

logger = logging.getLogger("Quart-WTF")

SUBMIT_METHODS = {"POST", "PUT", "PATCH", "DELETE"}

def _is_submitted() -> bool:
    """
    Consider the form submitted if there is an active request and
    the method is ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
    """
    return bool(request) and request.method in SUBMIT_METHODS

async def _get_formdata() -> Union[CombinedMultiDict, ImmutableMultiDict, MultiDict, None]:
    """
    Gets the formdata from the request.
    """
    req_files = await request.files
    req_form = await request.form

    if req_files:
        return CombinedMultiDict((req_files, req_form))
    if req_form:
        return req_form
    if request.is_json:
        req_json = await request.get_json()
        return ImmutableMultiDict(req_json)

    return None

def _get_config(
    value: Optional[Any],
    config_name: str,
    default: Optional[Any]=None,
    required: bool=True,
    message: str=CSRF_NOT_CONFIGURED
    ) -> Any:
    """
    Find config value based on provided value, Quart config, and default
    value.

    :param value: already provided config value
    :param config_name: Quart ``config`` key
    :param default: default value if not provided or configured
    :param required: whether the value must not be ``None``
    :param message: error message if required config is not found
    :raises KeyError: if required config is not found
    """
    if value is None:
        value = current_app.config.get(config_name, default)

    if required and value is None:
        raise RuntimeError(message)

    return value

def generate_csrf(
    secret_key: Optional[Any]=None,
    token_key: Optional[Any]=None
    ) -> Any:
    """
    Generate a CSRF token. The token is cached for a request, so multiple
    calls to this function will generate the same token.

    During testing, it might be useful to access the signed token in
    ``g.csrf_token`` and the raw token in ``session['csrf_token']``.

    :param secret_key: Used to securely sign the token. Default is
        ``WTF_CSRF_SECRET_KEY`` or ``SECRET_KEY``.
    :param token_key: Key where token is stored in session for comparison.
        Default is ``WTF_CSRF_FIELD_NAME`` or ``'csrf_token'``.
    """
    secret_key = _get_config(
        value=secret_key,
        config_name="WTF_CSRF_SECRET_KEY",
        default=current_app.secret_key,
        message=SECRET_KEY_REQUIRED
    )

    field_name = _get_config(
        value=token_key,
        config_name="WTF_CSRF_FIELD_NAME",
        default="csrf_token",
        message=FIELD_NAME_REQUIRED
    )

    if field_name not in g:
        serial = URLSafeTimedSerializer(secret_key, salt="wtf-csrf-token")

        if field_name not in session:
            session[field_name] = hashlib.sha1(os.urandom(64)).hexdigest()

        try:
            token = serial.dumps(session[field_name])
        except TypeError:
            session[field_name] = hashlib.sha1(os.urandom(64)).hexdigest()
            token = serial.dumps(session[field_name])

        setattr(g, field_name, token)

    return g.get(field_name)

def validate_csrf(
    data: Any,
    secret_key: Optional[Any]=None,
    time_limit: Optional[int]=None,
    token_key: Optional[Any]=None
    ) -> None:
    """"
    Check if the given data is a valid CSRF token. This compares the given
    signed token to the one stored in the session.

    :param data: The signed CSRF token to be checked.
    :param secret_key: Used to securely sign the token. Default is
        ``WTF_CSRF_SECRET_KEY`` or ``SECRET_KEY``.
    :param time_limit: Number of seconds that the token is valid. Default is
        ``WTF_CSRF_TIME_LIMIT`` or 3600 seconds (60 minutes).
    :param token_key: Key where token is stored in session for comparison.
        Default is ``WTF_CSRF_FIELD_NAME`` or ``'csrf_token'``.
    :raises ValidationError: Contains the reason that validation failed.
        Raises ``ValidationError`` with a specific error message rather than
        returning ``True`` or ``False``.
    """
    secret_key = _get_config(
        value=secret_key,
        config_name="WTF_CSRF_SECRET_KEY",
        default=current_app.secret_key,
        message=SECRET_KEY_REQUIRED
    )

    field_name = _get_config(
        value=token_key,
        config_name="WTF_CSRF_FIELD_NAME",
        default="csrf_token",
        message=FIELD_NAME_REQUIRED
    )

    time_limit = _get_config(
        value=time_limit,
        config_name="WTF_CSRF_TIME_LIMIT",
        default=3600,
        required=False
    )

    if not data:
        raise ValidationError(TOKEN_MISSING)

    if field_name not in session:
        raise ValidationError(SESSION_TOKEN_MISSING)

    serial = URLSafeTimedSerializer(secret_key, salt="wtf-csrf-token")

    try:
        token = serial.loads(data, max_age=time_limit)
    except SignatureExpired as error:
        raise ValidationError(TOKEN_EXPIRED) from error
    except BadData as error:
        raise ValidationError(TOKEN_INVALID) from error

    if not hmac.compare_digest(session[field_name], token):
        raise ValidationError(TOKEN_NO_MATCH)

def same_origin(current_uri: str, compare_uri: str) -> bool:
    """
    Determines if the request is from the same origin.

    :param current_uri: The current uri
    :param compare_uri: The uri to compare to the current uri.
    """
    current = urlparse(current_uri)
    compare = urlparse(compare_uri)

    return (
        current.scheme == compare.scheme
        and current.hostname == compare.hostname
        and current.port == compare.port
    )



"""
quart_wtf.const

Constant strings for CSRF.
"""
DEFAULT_ENABLED = True

DEFAULT_CHECK_DEFAULT = True

DEFAULT_CSRF_SECRET = None

DEFAULT_CSRF_FIELD_NAME = 'csrf_token'

DEFAULT_CSRF_TIME_LIMIT = 3600

DEFAULT_CSRF_HEADERS = ['X-CSRFToken', 'X-CSRF-Token']

DEFAULT_CSRF_SSL_STRICT = True

DEFAULT_SUBMIT_METHODS = ["POST", "PUT", "PATCH", "DELETE"]

CSRF_NOT_CONFIGURED = "CSRF is not configured.CSRF is not configured."

FIELD_NAME_REQUIRED = "A field name is required to used CSRF."

REFERRER_HEADER = "The referrer header is missing."

REFERRER_HOST = "The referrer does not match the host."

SECRET_KEY_REQUIRED = "A secret key is required to use CSRF."

SESSION_TOKEN_MISSING = "The CSRF session token is missing."

TOKEN_EXPIRED = "The CSRF token has expired."

TOKEN_INVALID = "The CSRF token is invalid"

TOKEN_MISSING = "The CSRF token is missing."

TOKEN_NO_MATCH = "The CSRF tokens do not match."

VALIDATION_FAILED = "CSRF validation failed."



"""
quart_wtf.meta

Defines the WTF meta and CSRF class for Quart-WTF.
"""
from typing import Any
from quart import current_app, g, session
from werkzeug.utils import cached_property
from wtforms import ValidationError
from wtforms.csrf.core import CSRF
from wtforms.meta import DefaultMeta

from .const import DEFAULT_ENABLED, DEFAULT_CSRF_FIELD_NAME, DEFAULT_CSRF_TIME_LIMIT
from .utils import logger, generate_csrf, validate_csrf

__all__ = ["QuartFormMeta"]

class _QuartFormCSRF(CSRF):
    meta = None

    def setup_form(self, form):
        """
        Setup the form for CSRF.
        """
        self.meta = form.meta

        return super().setup_form(form)

    def generate_csrf_token(self, csrf_token_field):

        return generate_csrf(
            secret_key=self.meta.csrf_secret,
            token_key=self.meta.csrf_field_name
        )

    def validate_csrf_token(self, form, field):

        if g.get("csrf_valid", False):
            # already validated by CSRFProtect.
            return

        try:
            validate_csrf(
                data=field.data,
                secret_key=self.meta.csrf_secret,
                time_limit=self.meta.csrf_time_limit,
                token_key=self.meta.csrf_field_name
            )
        except ValidationError as error:
            logger.info(error.args[0])
            raise

class QuartFormMeta(DefaultMeta):
    """
    Meta class for Quart specific subclass of WTForms.
    """
    csrf_class = _QuartFormCSRF
    csrf_context = session # not used, provided for custom CSRF class.

    @cached_property
    def csrf(self) -> bool:
        """
        Determines if CSRF is enabled.
        """
        return current_app.config.get("WTF_CSRF_ENABLED", DEFAULT_ENABLED)

    @cached_property
    def csrf_secret(self) -> Any:
        """
        CSRF secret key.
        """
        return current_app.config.get("WTF_CSRF_SECRET_KEY", current_app.secret_key)

    @cached_property
    def csrf_field_name(self) -> str:
        """
        CSRF field name.
        """
        return current_app.config.get('WTF_CSRF_FIELD_NAME', DEFAULT_CSRF_FIELD_NAME)

    @cached_property
    def csrf_time_limit(self) -> int:
        """
        CSRF time limit.
        """
        return current_app.config.get("WTF_CSRF_TIME_LIMIT", DEFAULT_CSRF_TIME_LIMIT)



"""
quart_wtf.csrf

The CSRF extension for Quart WTF.
"""
from typing import Any, Callable, Optional, Union
from quart import Quart, Blueprint, current_app, g, request
from werkzeug.exceptions import BadRequest
from wtforms import ValidationError

from .const import (DEFAULT_ENABLED, DEFAULT_CHECK_DEFAULT, DEFAULT_CSRF_FIELD_NAME,
                   DEFAULT_CSRF_HEADERS, DEFAULT_CSRF_TIME_LIMIT, DEFAULT_CSRF_SSL_STRICT,
                   DEFAULT_SUBMIT_METHODS, REFERRER_HEADER, REFERRER_HOST, VALIDATION_FAILED)
from .utils import logger, generate_csrf, validate_csrf, same_origin

__all__ = ["CSRFProtect"]

class CSRFProtect:
    """
    Enable CSRF protection globally for a Quart app.
    ::
        app = Quart(__name__)
        csrf = CSRFProtect(app)
    Checks the ``csrf_token`` field sent with forms, or the ``X-CSRFToken``
    header sent with JavaScript requests. Render the token in templates using
    ``{{ csrf_token() }}``.
    See the :ref:`csrf` documentation.
    """
    def __init__(self, app: Optional[Quart]=None) -> None:
        """
        Initialize the `CSRFProtect` class.
        """
        self._exempt_views = set()
        self._exempt_blueprints = set()

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Quart) -> None:
        """
        Initialize the `CSRFProtect` class with
        the `Quart` app.
        """
        app.extensions["csrf"] = self

        app.config.get("WTF_CSRF_ENABLED", DEFAULT_ENABLED)
        app.config.get("WTF_CSRF_CHECK_DEFAULT", DEFAULT_CHECK_DEFAULT)
        app.config.get("WTF_CSRF_METHODS", DEFAULT_SUBMIT_METHODS)
        app.config.get("WTF_CSRF_FIELD_NAME", DEFAULT_CSRF_FIELD_NAME)
        app.config.get("WTF_CSRF_HEADERS", DEFAULT_CSRF_HEADERS)
        app.config.get("WTF_CSRF_TIME_LIMIT", DEFAULT_CSRF_TIME_LIMIT)
        app.config.get("WTF_CSRF_SSL_STRICT", DEFAULT_CSRF_SSL_STRICT)

        app.jinja_env.globals["csrf_token"] = generate_csrf
        app.context_processor(lambda: {"csrf_token": generate_csrf})

        @app.before_request
        async def csrf_protect() -> None:
            if not current_app.config["WTF_CSRF_ENABLED"]:
                return
            if not current_app.config["WTF_CSRF_CHECK_DEFAULT"]:
                return
            if request.method not in current_app.config["WTF_CSRF_METHODS"]:
                return
            if not request.endpoint:
                return
            if request.blueprint in self._exempt_blueprints:
                return

            view = current_app.view_functions.get(request.endpoint)
            dest = f"{view.__module__}.{view.__name__}"

            if dest in self._exempt_views:
                return

            await self.protect()

    async def _get_csrf_token(self) -> Optional[Any]:
        """
        Gets the CSRF token.
        """
        # find the token in the form data.
        field_name = current_app.config["WTF_CSRF_FIELD_NAME"]
        form = await request.form
        base_token = form.get(field_name)

        if base_token:
            return base_token

        # if the form has a prefix, the name will be
        # {prefix}-csrf_token
        for key in form:
            if key.endswith(field_name):
                csrf_token = form[key]

                if csrf_token:
                    return csrf_token

        # find the token in the request headers
        for header_name in current_app.config["WTF_CSRF_HEADERS"]:
            csrf_token = request.headers.get(header_name)

            if csrf_token:
                return csrf_token

        return None

    async def protect(self) -> None:
        """
        Provides the CSRF protection for the app.
        """
        if request.method not in current_app.config["WTF_CSRF_METHODS"]:
            return

        try:
            validate_csrf(await self._get_csrf_token())
        except ValidationError as error:
            logger.info(error.args[0])
            self._error_response(error.args[0])

        if request.is_secure and current_app.config["WTF_CSRF_SSL_STRICT"]:
            if not request.referrer:
                self._error_response(REFERRER_HEADER)

            good_referrer = f"https://{request.host}/"

            if not same_origin(request.referrer, good_referrer):
                self._error_response(REFERRER_HOST)

        g.csrf_valid = True # Mark this request as CSRF valid.

    def exempt(self, view: Union[Blueprint, Callable, str]) -> Union[Blueprint, Callable, str]:
        """
        Mark a view or blueprint to be excluded from CSRF protection.
        ::
            @app.route('/some-view', methods=['POST'])
            @csrf.exempt
            async def some_view():
                ...
        ::
            bp = Blueprint(...)
            csrf.exempt(bp)
        """
        if isinstance(view, Blueprint):
            self._exempt_blueprints.add(view.name)
            return view

        if isinstance(view, str):
            view_location = view
        else:
            view_location = ".".join((view.__module__, view.__name__))

        self._exempt_views.add(view_location)
        return view

    def _error_response(self, reason: str) -> None:
        """
        Raises as a `CSRFError` with a specific reason.
        """
        raise CSRFError(reason)

class CSRFError(BadRequest):
    """Raise if the client sends invalid CSRF data with the request.
    Generates a 400 Bad Request response with the failure reason by default.
    Customize the response by registering a handler with
    :meth:`quart.Quart.errorhandler`.
    """

    description = VALIDATION_FAILED



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quart_wtf']

package_data = \
{'': ['*']}

install_requires = \
['WTForms>=3.0.1,<4.0.0', 'quart>=0.18.0,<0.19.0']

setup_kwargs = {
    'name': 'quart-wtf',
    'version': '0.0.1',
    'description': 'Simple integration of Quart and WTForms.',
    'long_description': '# Quart-WTF\n\n![Quart WTForms Logo](logos/logo.png)\n\nSimple integration of Quart and WTForms. Including CSRF and file uploading.\n',
    'author': 'Chris Rood',
    'author_email': 'quart.addons@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Quart-Addons/quart-uploads',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Quart-WTF Form
"""
from __future__ import annotations
import asyncio
from typing import Any, Coroutine, Dict, Optional, Union
from markupsafe import Markup
from wtforms import Field, Form, ValidationError
from wtforms.widgets import HiddenInput
from werkzeug.datastructures import MultiDict, CombinedMultiDict, ImmutableMultiDict

from .meta import QuartFormMeta
from .utils import _is_submitted, _get_formdata

_Auto = object()

class QuartForm(Form):
    """
    Quart specific subclass of WTForms :class:`~wtforms.form.Form`.
    To populate from submitted formdata use the ```.from_submit()``` class
    method to initialize the instance.
    """
    Meta = QuartFormMeta

    def __init__(
        self,
        formdata: Optional[Union[MultiDict, CombinedMultiDict, ImmutableMultiDict]]=None,
        obj: Optional[Any]=None,
        prefix: str="",
        data: Optional[Dict]=None,
        meta: Optional[Dict]=None,
        **kwargs
        ) -> None:
        """
        Initialize ``QuartForm`` class.
        """
        super().__init__(formdata, obj, prefix, data, meta, **kwargs)

    @classmethod
    async def create_form(
        cls,
        formdata: Union[object, MultiDict, CombinedMultiDict, ImmutableMultiDict]=_Auto,
        obj: Optional[Any]=None,
        prefix: str="",
        data: Optional[Dict]=None,
        meta: Optional[Dict]=None,
        **kwargs
        ) -> QuartForm:
        """
        This is a async method will create a new instance of ``QuartForm``.

        This method is primiarly used to intialize the class from submitted
        formdata from ``Quart.request``. If a request is a POST, PUT, PATCH,
        or DELETE method. The form will be intialized using the request. Otherwise
        it will initialized using the defaults. This is required since ``request.files``,
        ``request.form``, and ``request.json`` are coroutines and need to be called in an
        async manner.

        Also, if you are using ``quart_babel`` for translating components of this form, such
        as field labels. This method will call the lazy text as a coroutine, since ``quart_babel``
        use a coroutine to get the locale of of the user.

        """
        # Check if the formdata has not been passed to and if
        # if the form has been submitted. If it has been submitted
        # get it from `quart.request`.
        if formdata is _Auto:
            if _is_submitted():
                formdata = await _get_formdata()
            else:
                formdata = None

        return cls(formdata, obj, prefix, data, meta, **kwargs)

    async def _validate_async(self, validator: Coroutine, field: Field) -> bool:
        """
        Execute async validator.
        """
        try:
            await validator(self, field)
        except ValidationError as error:
            field.errors.append(error.args[0])
            return False
        return True

    async def validate(self, extra_validators: Optional[Dict]=None) -> bool:
        """
        Overload :meth:`validate` to handle custom async validators.
        """
        if extra_validators is not None:
            extra = extra_validators.copy()
        else:
            extra = {}

        async_validators = {}

        # use extra validators to check for StopValidation errors
        completed = []

        def record_status(form, field):
            completed.append(field.name)

        for name, field in self._fields.items():
            func = getattr(self.__class__, f"async_validate_{name}", None)
            if func:
                async_validators[name] = (func, field)
                extra.setdefault[name, []].append(record_status)

        # execute non-async validators
        success = super().validate(extra_validators=extra)

        # execute async validators
        tasks = [self._validate_async(*async_validators[name]) for name in \
            completed]
        async_results = await asyncio.gather(*tasks)

        # check results
        if False in async_results:
            success = False

        return success

    def is_submitted(self) -> bool:
        """
        Consider the form submitted if there is an active request and
        the method is ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        """
        return _is_submitted()

    async def validate_on_submit(self, extra_validators: Optional[Dict]=None) -> bool:
        """
        Call :meth:`validate` only if the form is submitted.
        This is a shortcut for ``form.is_submitted() and form.validate()``.
        """
        return self.is_submitted() and \
            await self.validate(extra_validators=extra_validators)

    def hidden_tag(self, *fields) -> Markup:
        """
        Render the form's hidden fields in one call.
        A field is considered hidden if it uses the
        :class:`~wtforms.widgets.HiddenInput` widget.
        If ``fields`` are given, only render the given fields that
        are hidden.  If a string is passed, render the field with that
        name if it exists.
        """
        def hidden_fields(fields):
            for field in fields:
                if isinstance(field, str):
                    field = getattr(self, field, None)

                if field is None or not isinstance(field.widget, HiddenInput):
                    continue

                yield field

        return Markup("\n".join(str(field) for field in hidden_fields(fields or self)))
